
#include <XTronical_ST7735.h> // Hardware-specific library
#include "GraphicsData.h"
#include "Graphics.h"
#include "Frogger.h"
#include "Levels.h"
#include <vector>
void InitLevel(Frogger_Class *Frogger,uint8_t LevelNumber)
{
  ClearAllObjects(Frogger);
  switch (LevelNumber)
  {
    case 1:InitLevel1(Frogger);break;
  }
    
}

void ClearAllObjects(Frogger_Class *Frogger)
{
  MovingObject_Class *ThisObj,*NextObj;

  ThisObj=Frogger->GameObjectListStart;
  while(ThisObj!=0)
  {
    NextObj=ThisObj->NextObj;
    delete(ThisObj);
    ThisObj=NextObj;
  }
  
  //Clear lady frog log pointers from lady frog log lane
  for(int i=0;i<5;i++)
    Frogger->LadyFrogLogs[i]=0;
  Frogger->LadyFrogIdx=0;
  Frogger->LadyFrogLog=0;
}


void InitLevel1(Frogger_Class *Frogger)
{  
  uint8_t i;
  std::vector<int16_t> Lane;
   
  Lane={40,74,112};                                                         VehicleLane(Frogger,CarLeft,-0.2,Lane,105,CarLeftWidth,CarLeftHeight);        //22s
  Lane={48,80,-15};                                                         VehicleLane(Frogger,BullDozer,0.24,Lane,97,BullDozerWidth,BullDozerHeight);   //18s
  Lane={48,80,114};                                                         VehicleLane(Frogger,PurpleCar,-0.3,Lane,90,PurpleCarWidth,PurpleCarHeight);   //14s
  Lane={-16};                                                               VehicleLane(Frogger,CarRight,0.4,Lane,80,CarRightWidth,CarRightHeight);       //10s
  Lane={60,104};                                                            VehicleLane(Frogger,Lorry,-0.6,Lane,73,LorryWidth,LorryHeight);               //7s
  Lane={0,1, 33,0, 65,0, 97,0};                                             TurtleLane(Frogger,-0.6,Lane,56,3);                                           //7s
  Lane={-29,18,58,98};                                                      LogLane(Frogger,0.24,Lane,LADY_FROG_LANE_Y,1,40);                                           //18s
  Lane={-52,12,76};                                                         LogLane(Frogger,1.2,Lane,41,4,32);                                            //3.5s
  Lane={19,1,  47,0,  75,0, 102,0};                                         TurtleLane(Frogger,-0.6,Lane,32,2);                                           //7s     
  Lane={-39,2,46,90};                                                       LogLane(Frogger,0.37,Lane,25,2,32);                                           //11s  
}

void VehicleLane(Frogger_Class *Frogger,const unsigned short *gfx,float Velocity,  std::vector<int16_t> & XData,uint8_t y,uint8_t width,uint8_t height)
{
  for(uint8_t i=0;i<XData.size();i++)
    AddObject(new MovingObject_Class(&Frogger->tft,gfx,width,height,XData[i],y,Velocity),Frogger);    
}

void LogLane(Frogger_Class *Frogger,float Velocity, std::vector<int16_t> & XData,uint8_t y,uint8_t Length,uint8_t DistToTravOffScr)
{
  Log_Class *Log;
  for(uint8_t i=0;i<XData.size();i++)    {
    Log=new Log_Class(&Frogger->tft,XData[i],y,Velocity,Length,DistToTravOffScr);
    AddObject(Log,Frogger);
    // If this is a log that can have a lady frog on then add to the special list
    if(y==LADY_FROG_LANE_Y)
    {
      Frogger->LadyFrogLogs[Frogger->LadyFrogIdx]=Log;  
      Frogger->LadyFrogIdx++;
    }
  }
}

void TurtleLane(Frogger_Class *Frogger,float Velocity, std::vector<int16_t> & XData,uint8_t y,uint8_t NumTurtlesPerGroup)
{
  // XData format :XData+0 -> x ord start for this group,  XData+1 ->diving or not -0 not 1 diving
  for(uint8_t i=0;i<XData.size();i+=2)    
  {
    AddObject(new Turtle_Class(&Frogger->tft,XData[i],y,Velocity,XData[i+1],NumTurtlesPerGroup),Frogger);      
  }
}


void Lane1(Frogger_Class *Frogger,float Velocity, std::vector<int16_t> & XData)
{
  for(uint8_t i=0;i<XData.size();i++)
    AddObject(new MovingObject_Class(&Frogger->tft,CarLeft,CarLeftWidth,CarLeftHeight,XData[i],105,Velocity),Frogger);  // 0.4 about 24 secs to travel across
    
}

void AddObject(MovingObject_Class *ThisObj,Frogger_Class *Frogger)
{
  // adds a moving object to the list of game objects
  
  if(Frogger->GameObjectListStart==0)              // empty at present, first entry
  {
    Frogger->GameObjectListStart=ThisObj;          // starts and ends with this entry
    Frogger->GameObjectListEnd=ThisObj;
  }
  else{
    Frogger->GameObjectListEnd->NextObj=ThisObj;
    ThisObj->PreviousObj=Frogger->GameObjectListEnd;
    Frogger->GameObjectListEnd=ThisObj;
  }
}

